#! /usr/bin/perl
#==============================================================================
# $Id: cleanmqlogs,v 1.19 2006/06/01 22:37:05 shelden Exp $
#==============================================================================
# This program can be used to remove linear logfiles no longer
# needed by MQSeries.
#==============================================================================
# ** WARNING ** WARNING ** WARNING ** WARNING ** WARNING **
#
# See README.html for warnings.  I think by now, everyone knows the
# risks.  And I've not had a report of the script ever screwing up
# it the 9+ years it's been in use.
#
# But summary of the above: caveat user.
#
#                               --Brian T. Shelden
#                                 Certified MQSeries Specialist
#                                 Shelden & Associates, Inc.
#                                 Austin, TX 78748-2319
#                                 shelden@shelden-associates.com
#==============================================================================
# Copyright (C) 1998-2002,2006 Shelden & Associates, Inc.
use Win32::TieRegistry;
use Getopt::Std;
use File::Basename;

if (!is_win32()) {
    $top        = "/var/mqm";         
    $log_top    = "/var/mqm/log";     
    $SEP        = "/";
}
else {
    $SEP        = "\\";
}
my($USAGE)      = join(''
                       , "usage: $0 [-z | -p | -u ]"
                       , " QueueManager [QueueManager...]\n"
                       );
#==============================================================================
my($qmgr, $fully_qualified_qmgrdir, @logfiles, @goners);
my($latest4media, $latest4restart, $latest, $file);
my($gzipem, $printem, $unlinkem);
my(%opt, %mq_qmgr_dirname);
#==============================================================================

&parse_argv;

die $USAGE
    unless @ARGV;

foreach $qmgr (@ARGV) {
    ($latest4media, $latest4restart) =
        &parse_error_messages($qmgr);
    print "$0: $qmgr: ", "Needed for Media Recovery: ", $latest4media, "\n";
    print "$0: $qmgr: ", "Needed for Qmgr Restart:   ", $latest4restart, "\n";
    undef $latest;
    if (&is_nearing_wraparound($qmgr)) {
        die <<"EOF";
$0: $qmgr is nearing a wraparound of linear logfiles.
    It would be reckless to clean them up automatically.  You will
    need to clean up the logfiles manually.
EOF
    }

    if (&mqlog_age_cmp($latest4media, $latest4restart) < 0) {
        $latest = $latest4media;
    }
    else {
        $latest = $latest4restart;
    }
    die(<<"EOF") unless $latest;
$0: Cannot determine latest logfiles for $qmgr.
$0: Perhaps its logfiles are circular?
EOF

    @goners = &can_be_deleted($qmgr, $latest);
    &cleanup($qmgr, @goners);
}
#==============================================================================
# This function returns the qmgr directory.
#==============================================================================
sub mq_qmgr_dirname {
    my($qmgr) = @_;
    my($mqsini) = join($SEP, $top, "mqs.ini");
    my($name, $prefix, $directory, $stanzaname);
    local($_, *INI);
    my($i);
    my(@lines);

    my($qmgr_hacked) = $qmgr;
    $qmgr_hacked =~ s/\./\!/g;

    undef $name;
    undef $prefix;
    undef $directory;

    if ($mq_qmgr_dirname{$qmgr}) {
        return $mq_qmgr_dirname{$qmgr};
    }

    if (&is_win32()) {
	my $top_of_qms = join('\\'
			      , "HKEY_LOCAL_MACHINE"
			      , "SOFTWARE"
			      , "IBM"
			      , "MQSeries"
			      , "CurrentVersion"
			      , "Configuration"
			      , "QueueManager"
			      );

	my $qms = $Registry-> {$top_of_qms};
	my $ibmkey;
	foreach $ibmkey (keys %$qms) {
	    $ibmkey =~ s,\\$,,;
	    foreach my $value ("Prefix", "Name", "Directory") {
		my $path = join('\\', $top_of_qms, $ibmkey, $value);
		my $x = $Registry-> {$path};
		if ($value eq "Prefix") {
		    $prefix = $x;
		}
		elsif ($value eq "Name") {
		    $name = $x;
		}
		elsif ($value eq "Directory") {
		    $directory = $x;
		}
	    }
	    my $dirname = join('\\'
			       , $prefix
			       , "qmgrs"
			       , $directory
			       );

	    $mq_qmgr_dirname{$name} = $dirname;

	    undef $name;
	    undef $prefix;
	    undef $directory;
        }
    }
    else {
        if (!open(INI, $mqsini)) {
            my $msg = "$0: error opening mqs.ini: $mqsini: $!\n";
            die $msg;
        }

        while(<INI>) {
#           print;
            if (/^(\w+):/) {
                # new stanza:
                $stanzaname = $1;
                $mq_qmgr_dirname{$name} = join($SEP, 
                                               $prefix, 
                                               "qmgrs", 
                                               $directory
                                               );
                undef $name;
                undef $prefix;
                undef $directory;
            }
            elsif (/^\s*Name=(.*)$/) {
                ($name = $1) if $stanzaname eq "QueueManager";
#               print $name, "\n";
            }
            elsif (/^\s*Prefix=(.*)$/) {
                $prefix = $1 if $stanzaname eq "QueueManager";
            }
            elsif (/^\s*Directory=(.*)$/) {
                $directory = $1 if $stanzaname eq "QueueManager";
            }
            else {
#               print "no match\n";
            }
        }
        $mq_qmgr_dirname{$name} = join($SEP, $prefix, "qmgrs", $directory);
        close(INI);
    }

    for (keys %mq_qmgr_dirname) {
        debug($_, " ", $mq_qmgr_dirname{$_}, "\n");
    }
    if (!$mq_qmgr_dirname{$qmgr}) {
        die("$0: no queue manager $qmgr found in "
	    , is_win32() ? "registry" : $mqsini
	    , "\n"
	    );
    }
    $mq_qmgr_dirname{$qmgr};
}
#==============================================================================
sub is_win32 {
    $^O eq 'MSWin32';
}
#==============================================================================
# This function returns the directory in which the recovery logs
# can be found.  
#==============================================================================
sub mq_logdir {
    my($qmgr) = @_;
    local(*INI, $_);
    my($qmini) = join($SEP, &mq_qmgr_dirname($qmgr), "qm.ini");
    my($rc);
    my(@lines);
    my($logtype);
    my($qmgr_hacked) = $qmgr;
    $qmgr_hacked =~ s/\./\!/g;

    if (&is_win32()) {
	### XXXX: Need to check for other logtypes here as well.
        my $path = join('\\'
                        , "HKEY_LOCAL_MACHINE"
                        , "SOFTWARE"
                        , "IBM"
                        , "MQSeries"
                        , "CurrentVersion"
                        , "Configuration"
                        , "QueueManager"
                        , $qmgr
                        , "Log"
                        , "LogPath"
                        );
        $rc = $Registry-> {$path};
    }
    else {
        if (!open(INI, $qmini)) {
            my $msg = "$0: error opening queue manager ini: $qmini: $!\n";
            die $msg;
        }
        while(<INI>) {
#           print;
            if (/^\s*LogPath=(.*)$/) {
                $rc = $1;
            }
            elsif (/^\s*LogType=(.*)\s*$/) {
                $logtype = $1;
                if ($logtype !~ /^LINEAR$/i) {
                    die("$0: the queue manager $qmgr "
                        , "uses $logtype logging, not linear.\n"
                        );
                }
            }
        }
        close INI;
    }
    join("", $rc, "active");
}
#==============================================================================
sub foo {
    my($x, $y) = @_;

    if ($x < $y) {
        return "< ";
    }
    elsif ($x > $y) {
        return "> ";
    }
    else {
        return "==";
    }
}
#==============================================================================
# Parse the /var/mqm/QMGR/errors logfiles in reverse time order.
# Well, ok; I'm cheating and assuming reverse *alpha* order
# will be reverse time order, but that's probably safe.
sub parse_error_messages {
    my($qmgr) = @_;
    my($file);
    my($latest4restart, $latest4media, $latest, $what, $msg);
    local($_);
    local(*FILE, *ERRDIR);
    my(@logfiles);
    my($fully_qualified_qmgrdir, $error_dir);

    $fully_qualified_qmgrdir = &mq_qmgr_dirname($qmgr);
    $error_dir = join($SEP, $fully_qualified_qmgrdir, "errors");

    local($/, $*);
    $/ = '-' x 10;
    $* = 1;

    opendir(ERRDIR, $error_dir)
        || die "$0: error opening directory $error_dir: $!\n";
    foreach $_ (readdir(ERRDIR)) {
        push(@logfiles, join($SEP, $error_dir, $_)) if /^AMQERR\d+.LOG$/;
    }

#    print "Here are the logfiles in question:\n", join("\n", @logfiles), "\n";

    foreach $file (reverse sort @logfiles) {
        open(FILE, $file)
            || die "$0: error opening $file: $!\n";
#       print "in $file:\n";
        while(<FILE>) {
#           print "Line ", $*, ": ", substr($_, 0, 60), "\n";
            s/\n/ /g;           # Makes the parsing easier...
            if (/(AMQ746[78]): .*(S\d+.LOG)/) {
                # Matched an entry like:
                # AMQ7467: The oldest log file required to start queue
                # manager FOO is S0000000.LOG.
                # or
                #
                # AMQ7468: The oldest log file required to perform
                # media recovery of queue manager FOO is S0000043.LOG.
                $msg = $1;
                $latest = $2;

                if ($msg =~ /7$/) {
                    # restart:
                    $latest4restart = $latest;
                    $what = "restart";
                }
                elsif ($msg =~ /8$/) {
                    # restart:
                    $latest4media = $latest;
                    $what = "media";
                }
#               print "The current latest for $what is ", $latest, "\n";
            }
        }
        close(FILE);
    }
    ($latest4media, $latest4restart);
}
#==============================================================================
sub can_be_deleted {
    my($qmgr) = @_;

    my($logdir, @logfiles, $file);
    my($latest_timestamp_m, $latest_timestamp_a, $latest_timestamp_c);
    my(@goners);
    local(*LOGDIR);
    local($_);

    $logdir = &mq_logdir($qmgr);
    opendir(LOGDIR, $logdir)
        || die "$0: error opening $logdir: $!\n";
    foreach $_ (readdir(LOGDIR)) {
        if (/^S\d+\.LOG$/) {
            my $f = join($SEP, $logdir, $_);
            push(@logfiles, $f);
            &debug("Pushing ", basename($f), " to \@logfiles\n");
        }
    }
    @logfiles = sort(@logfiles);

    $latest_timestamp_m = -M join($SEP, $logdir, $latest);
    $latest_timestamp_a = -A join($SEP, $logdir, $latest);
    $latest_timestamp_c = -C join($SEP, $logdir, $latest);

    foreach $file (@logfiles) {
#       printf("%s M(%3.4f %s %3.4f) A(%3.4f %s %3.4f) C(%3.4f %s %3.4f)\n",
        my(@x) = (
                  basename($file),
                  -M $file,
                  &foo(-M $file, $latest_timestamp_m),
                  $latest_timestamp_m,
                  -A $file,
                  &foo(-A $file, $latest_timestamp_a),
                  $latest_timestamp_a,
                  -C $file,
                  &foo(-C $file, $latest_timestamp_c),
                  $latest_timestamp_c
              );


        #print "This is a bug; \"older\" means a \"a number less than\"\n";
#       if ($latest_timestamp_a < -A $file) {
        if (&mqlog_age_cmp($latest, $file) > 0) {
#           print $file, " can go, it is older than ", $latest, "\n";
            push(@goners, $file);
        }
    }
    @goners;
}
#==============================================================================
# This cmp function returns a negative number if $a is older than
# $b, zero if $a == $b, and a positive number if $a is newer than
# $b.  We don't use timestamps because when a message spans
# more than one logfile (which happens whenever the message is near
# or bigger than the logfile size), the later file (by number)
# has an *earlier* timestamp.  :-(
# So, we *must* go by name, not by timestamp, and deal with
# wraparound (S9999999.LOG followed by S0000000.LOG) manually.
sub mqlog_age_cmp {
    my($a, $b) = @_;
    my($basea, $baseb) = (basename($a), basename($b));

    $basea cmp $baseb;
}
#==============================================================================
sub is_nearing_wraparound {
    my($qmgr) = @_;
    my($logdir) = &mq_logdir($qmgr);
    my(@logfiles);
    local($_);

    opendir(LOGDIR, $logdir)
        || die "$0: error opening directory $logdir: $!\n";
    foreach $_ (readdir(LOGDIR)) {
        if (/^S9999.*.LOG$/) {
            # We're within a 1000 of rolling over.  Return true:
            closedir(LOGDIR);
            return 1;
        }
    }
    closedir(LOGDIR);
    0;
}
#==============================================================================
sub parse_argv {
#    my($opt_z, $opt_Z, $opt_p, $opt_u);

    no strict "vars";
    undef $opt_z;               # defeats -w
    undef $opt_p;               # defeats -w
    undef $opt_u;               # defeats -w
    undef $opt_t;               # defeats -w
    undef $opt_l;               # defeats -w

    &getopts("nt:l:zZpu") || die $USAGE;
    $opt_z && ($gzipem       = 1);
    $opt_p && ($printem      = 1);
    $opt_u && ($unlinkem     = 1);
    $opt_t && ($top          = $opt_t);
    $opt_l && ($log_top      = $opt_l);
    use strict;

    if (!$gzipem && !$printem && !$unlinkem) {
        $gzipem = 1;
    }
    (
     ($gzipem  && !$printem && !$unlinkem) ||
     (!$gzipem && !$printem && !$unlinkem) ||
     (!$gzipem &&  $printem && !$unlinkem) ||
     (!$gzipem && !$printem &&  $unlinkem)
     )
        || die "$0: -z, -p, and -u are mutually exclusive\n";

#    print("Action is: ",
#         $gzipem       ? "gzip"        :
#         $printem      ? "print"       :
#         "unlink",
#         " on qmgrs: ",
#         join(" ", @ARGV),
#         "\n");
#   exit;
}
#==============================================================================
sub cleanup {
    my($qmgr, @goners) = @_;
    my($action) =
        $gzipem ? "gzip" :
            $printem ? "print" :
                $unlinkem ? "unlink" :
                    "unknown";

    die "$0: internal error: unknown action\n" if ($action eq "unknown");
    foreach $file (@goners) {
        print "About to ", $action, " ", $qmgr, "'s ", 
            basename($file), "\n";

        if ($gzipem) {
            chdir(dirname($file))
                || die "$0: can't cd to ", dirname($file), ": $!\n";
            !system($GZIP, basename($file))
                || die "$0: can't gzip $file: $!\n";
        }
        elsif ($unlinkem) {
            unlink($file) || die "$0: can't unlink $file: $!\n";
        }
        # printem is taken care of above.
    }
}
#==============================================================================
sub debug {
    my($msg) = join(' ', @_);
    if (0) {
        print $msg;
    }
}
#==============================================================================
